import re

from ...helpers.config import build_config, core_config
from ...helpers.debian import apt_install
from ...helpers.fs import (
    insert_or_replace,
    render_template,
    set_file_contents,
    set_file_contents_from_template,
)
from ...helpers.network import wireguard_ip
from ...helpers.systemd import systemd_set
from .. import wireguard
from . import check_master
from .common import get_master_ip

options = {}


def dependencies():
    return [wireguard, check_master]


def run():
    apt_install(["redis-sentinel", "redis-server"])
    LOCAL = build_config(core_config())
    local_ip = wireguard_ip()
    master_ip = get_master_ip()
    server_changes = set_file_contents_from_template(
        "/etc/redis/redis.conf",
        "redis.conf.j2",
        WIREGUARD_IP=local_ip,
        IS_MASTER=local_ip == master_ip,
        MASTER_IP=master_ip,
        REDIS_PASSWORD=LOCAL["REDIS_PASSWORD"],
    )
    sentinel_changes = False
    # We need to have all the lines we've added, but sentinel messes with things and might add others
    rendered = render_template(
        "redis-sentinel.conf.j2",
        WIREGUARD_IP=local_ip,
        MASTER_IP=master_ip,
        QUORUM_REQUIRED=options.get("quorum", "2"),
        REDIS_PASSWORD=LOCAL["REDIS_PASSWORD"],
    )
    current = open("/etc/redis/sentinel.conf").read()
    if "Example sentinel.conf" in current:
        # original Debian one, replace
        sentinel_changes = set_file_contents("/etc/redis/sentinel.conf", rendered)
    else:
        for line in rendered.split("\n"):
            if line.strip() == "":
                continue
            if line.startswith("bind"):
                patt = re.compile("bind.+")
            elif line.startswith("user default on"):
                continue
            elif line.find("mymaster") != -1:
                patt = re.compile(line[: line.find("mymaster")] + "mymaster.+")
            else:
                patt = re.compile(" ".join(line.split(" ")[:-1]) + ".+")
            sentinel_changes = (
                insert_or_replace("/etc/redis/sentinel.conf", patt, line)
                or sentinel_changes
            )
    systemd_set("redis-server", enabled=True, running=True, restart=server_changes)
    systemd_set("redis-sentinel", enabled=True, running=True, restart=sentinel_changes)



import json
import os
import pathlib
from typing import Any, Dict, List, Optional, cast

import jinja2
import yaml
from mergedeep import merge
from typing_extensions import NotRequired, TypedDict

from paracrine import is_dry_run


class ServerDict(TypedDict):
    name: str
    count: int
    ssh_hostname: str
    ssh_port: int
    ssh_key: str
    ssh_user: str
    wireguard_ip: NotRequired[str]


class InventoryDict(TypedDict):
    environment: str
    data_path: str
    servers: List[ServerDict]


class ConfigDict(TypedDict):
    environments: Dict[str, object]
    common: NotRequired[Dict[str, object]]


_jinja_env = None
data = None

CONFIG_NAME = "config.yaml"


def jinja_env():
    assert _jinja_env is not None, "Need to run set_data first!"
    return _jinja_env


def host():
    assert data is not None
    return data["host"]


def config() -> InventoryDict:
    assert data is not None
    return data["config"]


def data_files():
    assert data is not None
    return data["data"]


def get_config_keys():
    assert data is not None
    return data["configs"].keys()


def get_config_file(fname: str) -> str:
    if fname not in get_config_keys():
        raise KeyError(f"Can't find {fname}. We have: {sorted(get_config_keys())}")

    assert data is not None
    return data["configs"][fname]


def core_config():
    return yaml.safe_load(get_config_file(CONFIG_NAME))


def set_data(new_data: Dict[str, Any]) -> None:
    loader = jinja2.DictLoader(new_data["templates"])
    global _jinja_env, data
    _jinja_env = jinja2.Environment(
        loader=loader, undefined=jinja2.StrictUndefined, keep_trailing_newline=True
    )
    data = new_data


return_data: Dict[object, object] = {}


def clear_return_data() -> None:
    global return_data
    return_data = {}


def add_return_data(new_data: Dict[str, Any]) -> None:
    global return_data
    merge(return_data, new_data)


def get_return_data() -> Dict[object, object]:
    global return_data
    return return_data


def add_folder_to_config(configs, folder, shortname=None, filter=None, prefix=""):
    if not os.path.exists(folder):
        print("Skipping %s from config as doesn't exist" % folder)
        return
    for f in os.listdir(folder):
        if filter is not None:
            if not filter(f):
                continue
        if shortname is not None:
            key = os.path.join(shortname, f)
            full = os.path.join(folder, f)
        else:
            full = key = os.path.join(folder, f)
        if os.path.isfile(full):
            if prefix != "":
                if "/" in key:
                    parts = key.split("/")
                    key = "/".join(parts[:-1]) + "/" + prefix + parts[-1]
                else:
                    key = prefix + key
            configs[key] = open(full).read()
        else:
            if prefix != "":
                prefix += "/"
            add_folder_to_config(
                configs,
                full,
                shortname=shortname,
                filter=filter,
                prefix=prefix + f + "/",
            )


inventory: Optional[InventoryDict] = None
inventory_directory: Optional[str] = None


def set_config(inventory_path: str) -> None:
    global inventory, inventory_directory
    inventory_directory = os.path.dirname(inventory_path)
    inventory = yaml.safe_load(open(inventory_path))


def get_config() -> InventoryDict:
    if inventory is None:
        assert data is not None
        return data["inventory"]
    return inventory


def config_path(shortname: bool = False) -> str:
    if inventory is None or shortname:
        return "configs"
    else:
        dp = data_path()
        assert dp is not None
        return os.path.join(dp, "configs")


def path_to_config_file(name: str) -> pathlib.Path:
    if name.find("~") != -1:
        return pathlib.Path(name).expanduser()
    return pathlib.Path(config_path()).joinpath(name)


def data_path():
    if inventory is None:
        return None
    assert inventory_directory is not None
    return os.path.join(inventory_directory, inventory["data_path"])


def environment():
    if inventory is None:
        assert data is not None
        return data["environment"]
    return inventory["environment"]


def servers() -> List[ServerDict]:
    try:
        return get_config()["servers"]
    except NotADirectoryError:
        return core_config()["servers"]


def walk(path):
    for p in pathlib.Path(path).iterdir():
        if p.is_dir():
            yield from walk(p)
            continue
        yield p.resolve()


def create_data(server: Optional[ServerDict] = None):
    config = get_config()
    templates = {}
    template_paths = [
        pathlib.Path("templates"),
        pathlib.Path(__file__).parent.parent.joinpath("templates"),
    ]
    for template_path in template_paths:
        if not template_path.exists():
            continue
        for path in walk(template_path):
            templates[path.name] = path.open("r").read()

    data = {}
    data_paths = [
        pathlib.Path("data"),
        pathlib.Path(__file__).parent.joinpath("data"),
    ]
    for data_path in data_paths:
        if not data_path.exists():
            continue
        data_path = data_path.absolute()
        for path in walk(data_path):
            try:
                local_path = path.relative_to(data_path).as_posix()
            except ValueError:
                # Symlink outside of local folder
                continue
            data[local_path] = path.open("rb").read()

    configs = {
        CONFIG_NAME: open(CONFIG_NAME).read(),
    }
    add_folder_to_config(
        configs,
        config_path(),
        shortname="configs",
        filter=lambda f: not f.startswith("."),
    )

    return {
        "templates": templates,
        "host": server,
        "config": config,
        "configs": configs,
        "environment": environment(),
        "inventory": get_config(),
        "data": data,
    }


def network_config_file(name, shortname=False):
    return config_path(shortname) + "/networks-{name}".format(name=name)


def network_config(name):
    try:
        return json.loads(get_config_file(network_config_file(name, shortname=True)))
    except KeyError:
        if not is_dry_run():
            raise
        return {}


def other_config_file(name, shortname=False):
    return config_path(shortname) + "/other-{name}".format(name=name)


def other_config(name):
    return json.loads(get_config_file(other_config_file(name, shortname=True)))


def other_self_config():
    return other_config(host()["name"])


def build_config(config: ConfigDict) -> Dict[str, object]:
    env = environment()
    LOCAL = cast(Dict[str, object], config["environments"][env])
    common = cast(Dict[str, object], config.get("common", {}))
    ret = cast(Dict[str, object], dict(**common))
    ret.update(LOCAL)
    return ret


def local_config() -> ConfigDict:
    return yaml.safe_load(open(CONFIG_NAME).read())


def local_server():
    local_hostname = host()["name"]
    for server in servers():
        name = server["name"]
        if name == local_hostname:
            return server

    raise Exception(f"Cannot find {local_hostname}")


def in_docker():
    return os.path.exists("/.dockerenv")



import argparse
import json
import logging
import os
from typing import Any, Callable, Dict, List

from mergedeep import merge
from mitogen.core import Error, StreamError
from mitogen.parent import Context, EofError, Router
from mitogen.utils import run_with_router
from retry.api import retry_call

from paracrine import DRY_RUN_ENV

from .deps import (
    Modules,
    TransmitModules,
    makereal,
    maketransmit,
    maketransmit_single,
    runfunc,
)
from .helpers.config import (
    create_data,
    get_config,
    other_config_file,
    path_to_config_file,
    set_config,
    set_data,
)
from .helpers.fs import set_file_contents
from .runners import core


def decode(info):
    for k in list(info.keys()):
        if isinstance(k, bytes):
            info[k.decode()] = info[k].decode()
            del info[k]


ssh_cache: Dict[str, Context] = {}


def clear_ssh_cache():
    global ssh_cache
    ssh_cache = {}


def main(router: Router, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Dict:
    config = get_config()
    calls = []
    wg = core.is_wireguard()
    data = None
    for server in config["servers"]:
        assert isinstance(server, Dict)
        hostname = (
            server["wireguard_ip"]
            if wg and "wireguard_ip" in server
            else server["ssh_hostname"]
        )
        port = 22 if wg else server.get("ssh_port", 22)
        cache_key = f"{hostname}-{port}"
        if cache_key not in ssh_cache:
            key_path = path_to_config_file(server["ssh_key"])
            if not os.path.exists(key_path):
                raise Exception(f"Can't find ssh key {key_path}")
            username = server["ssh_user"]

            try:
                connect = retry_call(
                    router.ssh,
                    exceptions=EofError,
                    tries=3,
                    fkwargs={
                        "hostname": hostname,
                        "port": port,
                        "username": username,
                        "identity_file": key_path,
                        "check_host_keys": "accept",
                        "python_path": "python3",
                    },
                )
            except StreamError:
                print(
                    "Exception while trying to login to %s@%s:%s"
                    % (username, hostname, port)
                )
                raise

            if username != "root":
                sudo = router.sudo(
                    via=connect, python_path="python3", preserve_env=True
                )
                ssh_cache[cache_key] = sudo
            else:
                ssh_cache[cache_key] = connect
        data = create_data(server=server)
        calls.append(ssh_cache[cache_key].call_async(func, data, *args, **kwargs))

    infos = []
    errors = []
    for call in calls:
        try:
            info = call.get().unpickle()
            if info is not None:
                decode(info)
            infos.append(info)
        except Error as e:
            print("Got error", e)
            errors.append(e)

    if len(errors) > 0:
        raise Exception(errors)

    return {"infos": infos, "data": data}


def do(data, transmitmodules: TransmitModules, name: str, dry_run: bool):
    os.environ[DRY_RUN_ENV] = str(dry_run)
    set_data(data)
    modules = makereal(transmitmodules)
    return runfunc(modules, name)


def internal_runner(
    router: Router,
    modules: Modules,
    local_func: str,
    run_func: str,
    parse_func: str,
    dry_run: bool,
) -> None:
    for module in modules:
        runfunc([module], local_func)
        infos = main(router, do, maketransmit([module]), run_func, dry_run)
        for info in infos["infos"]:
            os.environ[DRY_RUN_ENV] = str(dry_run)
            runfunc([module], parse_func, info, infos["data"])
            for module_name in info:
                for per_node in info[module_name]:
                    if not isinstance(per_node, Dict):
                        continue
                    if "selector" not in per_node:
                        continue
                    config_path = other_config_file("selectors.json")
                    if os.path.exists(config_path):
                        selector_config = json.load(open(config_path))
                    else:
                        selector_config = {}
                    merge(selector_config, per_node["selector"])
                    selector_config = dict(sorted(selector_config.items()))
                    set_file_contents(
                        config_path, json.dumps(selector_config, indent=2)
                    )


def generate_dependencies(modules: Modules):
    tree = {}
    mapping = {}
    checked = []
    needs_dependencies = list(modules) + [core]
    modules = []
    while len(needs_dependencies) > 0:
        check = needs_dependencies.pop()
        checked.append(check)
        key = str(maketransmit_single(check))
        tree[key] = []
        mapping[key] = check
        new_dependencies = runfunc([check], "dependencies")
        for item in new_dependencies.values():
            for new_dependency in item[0]:
                tree[key].append(new_dependency)
                if new_dependency in needs_dependencies or new_dependency in checked:
                    continue
                needs_dependencies.append(new_dependency)

    count = 0
    while len(checked) > 0:
        count += 1
        if count == 100:
            raise Exception((modules, checked))
        for key, value in tree.items():
            mapped = mapping[key]
            if mapped in modules:
                continue
            for dep in value:
                if dep not in modules:
                    break
            else:
                modules.append(mapped)
                checked.remove(mapped)

    return modules


def run(args: List[str], modules: Modules):
    logging.basicConfig()
    logging.root.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(prog="paracrine")
    parser.add_argument("-i", "--inventory-path", dest="inventory_path", required=True)
    parser.add_argument("-a", "--apply", default=False, action="store_true")
    parsed_args = parser.parse_args(args)
    set_config(parsed_args.inventory_path)

    modules = generate_dependencies(modules)

    print("Running:")
    for module in maketransmit(modules):
        print(f"* {module}")
    print("")

    run_with_router(
        internal_runner, modules, "local", "run", "parse_return", not parsed_args.apply
    )



from pathlib import Path

from paracrine import dry_run_safe_read

from ...helpers.config import config, get_config_file, host, in_docker
from ...helpers.fs import set_file_contents_from_template
from ...helpers.network import external_ip, wireguard_ip
from ...helpers.systemd import systemd_set
from . import bootstrap
from .common import private_key_file, public_key_path


def dependencies():
    return [bootstrap]


def setup(name: str = "wg0", ip: str = "192.168.2.1", netmask: int = 24):
    peers = {}
    for h in config()["servers"]:
        if h["name"] == host()["name"]:
            continue
        public_key = get_config_file(public_key_path(h["name"]))
        peers[h["name"]] = {
            "public_key": public_key,
            "endpoint": "%s:51820" % external_ip(h),
            "peer_addr": h.get("wireguard_ip", "<unknown>"),
        }

    conf_change = set_file_contents_from_template(
        f"/etc/wireguard/{name}.conf",
        "wg.conf.j2",
        PRIVATE_KEY=dry_run_safe_read(
            Path(private_key_file), "fake private key"
        ).strip(),
        PEERS=peers,
        IP=ip,
        NETMASK=netmask,
    )

    if not in_docker():
        systemd_set(f"wg-quick@{name}", enabled=True, restart=conf_change)


def run():
    setup(ip=wireguard_ip())



import logging
from pathlib import Path

import requests
import urllib3

from ...helpers.config import get_config_file, get_config_keys, in_docker
from ...helpers.fs import (
    download_and_unpack,
    make_directory,
    set_file_contents,
    set_file_contents_from_template,
    set_mode,
)
from ...helpers.network import wireguard_ips
from ...helpers.systemd import systemctl_daemon_reload, systemd_set
from ...helpers.users import adduser
from .. import wireguard
from . import certs
from .common import (
    CERTS_DIR,
    HOME_DIR,
    USER,
    binary_path,
    cockroach_hash,
    cockroach_url,
    local_node_ip,
)

options = {}

urllib3.disable_warnings()


def dependencies():
    return [certs, wireguard]


def run():
    unpacked = download_and_unpack(
        cockroach_url,
        cockroach_hash,
    )
    cockroach_path = Path(unpacked["dir_name"]).joinpath(binary_path)
    adduser(USER, HOME_DIR)
    make_directory(CERTS_DIR)
    file_changes = False
    for fname in get_config_keys():
        if "cockroach-certs" not in fname:
            continue
        new_fname = CERTS_DIR.joinpath(
            fname.replace("configs/cockroach-certs/", "")
        ).as_posix()
        if local_node_ip() in fname:
            new_fname = new_fname.replace(local_node_ip(), "node")
        elif "192.168" in new_fname:  # wireguard
            continue
        file_changes = (
            set_file_contents(new_fname, get_config_file(fname), owner="cockroach")
            or file_changes
        )
        file_changes = set_mode(new_fname, "700") or file_changes

    COCKROACH_PORT = options.get("COCKROACH_PORT", 26257)
    SQL_PORT = options.get("SQL_PORT", 26258)
    service_file_changes = set_file_contents_from_template(
        "/etc/systemd/system/cockroach.service",
        "cockroach.service.j2",
        COCKROACH_PATH=cockroach_path,
        HOME_DIR=HOME_DIR,
        USER=USER,
        CERTS_DIR=CERTS_DIR,
        WIREGUARD_IP=local_node_ip(),
        HTTP_PORT=options.get("HTTP_PORT", 8080),
        COCKROACH_PORT=COCKROACH_PORT,
        SQL_PORT=SQL_PORT,
        HOST_LIST=",".join(
            [f"{ip}:{COCKROACH_PORT}" for ip in wireguard_ips().values()]
        ),
    )
    if service_file_changes:
        systemctl_daemon_reload()

    needs_restart = file_changes or service_file_changes
    if needs_restart is False:
        # Check it's up and doesn't need a restart anyways
        try:
            resp = requests.get(
                f"https://{local_node_ip()}:9080/_admin/v1/health", verify=False
            )
            if resp.status_code != 200:
                logging.warning(
                    f"Cockroach status was {resp.status_code} so restarting"
                )
                needs_restart = True
        except requests.exceptions.ConnectionError:
            logging.warning("Can't connect to Cockroach health endpoint, so restarting")
            needs_restart = True

    # FIXME: Can't make this work in Docker
    if not in_docker():
        systemd_set(
            "cockroach",
            enabled=True,
            running=True,
            restart=needs_restart,
        )



from ...helpers.config import config_path

wg_config = "/etc/wireguard"
private_key_file = f"{wg_config}/privatekey"
public_key_file = f"{wg_config}/publickey"


def public_key_path(name: str) -> str:
    return config_path() + f"/wireguard-public-{name}.key"



from paracrine.deps import Modules, maketransmit
from paracrine.runner import generate_dependencies
from paracrine.services import cockroachdb, wireguard


def test_deps_ordering():
    modules: Modules = [wireguard, cockroachdb]
    modules = generate_dependencies(modules)

    assert maketransmit(modules) == [
        "paracrine.runners.core",
        "paracrine.services.cockroachdb.certs",
        "paracrine.services.wireguard.bootstrap",
        "paracrine.services.wireguard.core",
        "paracrine.services.wireguard",
        ("paracrine.services.cockroachdb.node", {}),
        ("paracrine.services.cockroachdb.init", {}),
        "paracrine.services.cockroachdb",
    ]



import json
import pathlib

from ..helpers.config import config_path, get_config_file
from ..helpers.fs import run_command, set_file_contents


def local():
    aws_path = pathlib.Path(config_path()).joinpath("other-aws")
    if not aws_path.exists():
        set_file_contents(
            aws_path,
            json.dumps(
                {
                    "access_key": run_command(
                        "aws configure get aws_access_key_id"
                    ).strip(),
                    "secret_key": run_command(
                        "aws configure get aws_secret_access_key"
                    ).strip(),
                },
                indent=2,
            ),
        )


def get_env_with_creds():
    aws_conf = json.loads(get_config_file("configs/other-aws"))
    return f"env AWS_ACCESS_KEY_ID={aws_conf['access_key']} AWS_SECRET_ACCESS_KEY={aws_conf['secret_key']}"



import json
import os
import re

from ...helpers.config import build_config, core_config, other_config, other_config_file
from ...helpers.debian import apt_is_installed
from ...helpers.fs import run_command
from ...helpers.network import wireguard_ip
from ...runners.core import wireguard_ip_for_machine_for

MASTER_FILE = "masters.json"


def get_existing_masters(local):
    if local:
        master_file = other_config_file(MASTER_FILE)
        if os.path.exists(master_file):
            return json.load(open(master_file))
        else:
            return []
    else:
        try:
            return other_config("masters.json")
        except KeyError:
            return []


def get_master_ip():
    existing_masters = get_existing_masters(False)
    if len(existing_masters) > 0:
        if all([master == existing_masters[0] for master in existing_masters]):
            # everyone agrees, awesome
            return existing_masters[0]
        rated = {}
        for master in existing_masters:
            if master not in rated:
                rated[master] = 1
            else:
                rated[master] += 1
        highest = sorted(rated.values(), reverse=True)[0]
        best_masters = [k for k in rated.keys() if rated[k] == highest]
        if len(best_masters) != 1:
            raise Exception(
                f"Can't choose, because {len(best_masters)} with {highest} entries in {best_masters}. See also {existing_masters}"
            )

        return best_masters[0]

    if not apt_is_installed("redis-tools"):
        return wireguard_ip_for_machine_for("redis-master")

    LOCAL = build_config(core_config())
    local_ip = wireguard_ip()
    try:
        output = run_command(
            f"redis-cli -a {LOCAL['REDIS_PASSWORD']} -h {local_ip} info replication"
        )
    except AssertionError:
        # redis isn't up
        return wireguard_ip_for_machine_for("redis-master")
    if "role:master" in output and ",port=6379,state=online" in output:
        master_ip = local_ip
    else:
        master_host_match = re.search("master_host:(.+)", output)
        if master_host_match is not None:
            master_ip = master_host_match.groups()[0].strip()
        else:
            master_ip = wireguard_ip_for_machine_for("redis-master")

    return master_ip



from pathlib import Path

from ..deps import Modules
from ..helpers.config import build_config, core_config, get_config_file, local_config
from ..helpers.debian import apt_install
from ..helpers.fs import (
    download_and_unpack,
    link,
    make_directory,
    run_command,
    run_with_marker,
    set_file_contents,
    set_file_contents_from_template,
)
from ..helpers.systemd import link_service, systemd_set
from ..helpers.users import adduser
from ..runners import certs
from . import postgresql

# FIXME: Do soapbox from https://gitlab.com/soapbox-pub/soapbox/-/jobs/3371276281/artifacts/download
# Unzipped with "unzip soapbox.zip -d instance" and then moved instance/static to /var/lib/pleroma
# https://docs.soapbox.pub/frontend/installing/#install-soapbox


def dependencies() -> Modules:
    LOCAL = build_config(local_config())
    return [
        postgresql,
        (
            certs,
            {"hostname": LOCAL["PLEROMA_HOST"], "email": LOCAL["PLEROMA_EMAIL"]},
        ),
    ]


def run():
    LOCAL = build_config(core_config())
    adduser("pleroma", home_dir="/opt/pleroma")
    make_directory("/opt/pleroma", owner="pleroma")
    # Taken from https://git.pleroma.social/pleroma/pleroma/-/pipelines?page=1&scope=branches&ref=stable using amd64:archive
    res = download_and_unpack(
        "https://git.pleroma.social/pleroma/pleroma/-/jobs/234433/artifacts/download?file_type=archive",
        "8ef0bea62671d39e60f9e08d13109a4c332c552a1f855184063353987d46c84a",
        "pleroma-2.5.2.zip",
        "/opt/pleroma-2.5.2",
    )
    release_changed = res["changed"]

    if release_changed:
        run_command("cp -R /opt/pleroma-2.5.2/release/* /opt/pleroma")
        run_command("chown -R pleroma /opt/pleroma")

    make_directory("/var/lib/pleroma/uploads", owner="pleroma")
    make_directory("/var/lib/pleroma/static", owner="pleroma")
    set_file_contents_from_template("/var/lib/pleroma/static/robots.txt", "robots.txt")
    make_directory("/etc/pleroma", owner="pleroma")
    config_changes = set_file_contents_from_template(
        "/etc/pleroma/config.exs", "config.exs.j2", ignore_changes=False, **LOCAL
    )
    db_changes = set_file_contents_from_template(
        "/opt/pleroma/setup_db.psql", "setup_db.psql.j2", ignore_changes=False, **LOCAL
    )

    run_with_marker(
        "/opt/pleroma/setup_db.marker",
        'su postgres -s $SHELL -lc "psql -f /opt/pleroma/setup_db.psql"',
        deps=["/opt/pleroma/setup_db.psql"],
        force_build=release_changed and db_changes,
        run_if_command_changed=False,
    )
    run_with_marker(
        "/opt/pleroma/migrate.marker",
        'su pleroma -s $SHELL -lc "./bin/pleroma_ctl migrate"',
        directory="/opt/pleroma",
        force_build=release_changed,
    )

    link_service("/opt/pleroma/installation/pleroma.service")
    systemd_set(
        "pleroma", enabled=True, running=True, restart=release_changed or config_changes
    )

    cert_dir = Path("/opt/letsencrypt")
    nginx_changes = make_directory(str(cert_dir))
    nginx_changes = (
        set_file_contents(
            cert_dir.joinpath("fullchain.pem"),
            get_config_file(f"configs/other-fullchain-{LOCAL['PLEROMA_HOST']}"),
        )
        or nginx_changes
    )
    nginx_changes = (
        set_file_contents(
            cert_dir.joinpath("privkey.pem"),
            get_config_file(f"configs/other-privkey-{LOCAL['PLEROMA_HOST']}"),
        )
        or nginx_changes
    )
    nginx_changes = (
        set_file_contents(
            cert_dir.joinpath("options-ssl-nginx.conf"),
            get_config_file("configs/other-ssl-options"),
        )
        or nginx_changes
    )

    apt_install(["nginx"])

    nginx_changes = (
        set_file_contents_from_template(
            "/etc/nginx/sites-available/pleroma.conf",
            "pleroma.nginx.j2",
            PLEROMA_HOST=LOCAL["PLEROMA_HOST"],
            DUMMY_CERTS=certs.get_dummy_certs(),
        )
        or nginx_changes
    )
    nginx_changes = (
        link(
            "/etc/nginx/sites-enabled/pleroma.conf",
            "/etc/nginx/sites-available/pleroma.conf",
        )
        or nginx_changes
    )

    if nginx_changes:
        systemd_set("nginx", reloaded=True)



import subprocess
import sys

from ..helpers.config import get_config, path_to_config_file, set_config


def ssh_server(server, run="", as_root=False):
    if run != "" and as_root and server["ssh_user"] != "root":
        run = f"sudo {run}"
    command = f"ssh {server['ssh_user']}@{server['ssh_hostname']} \
        -i {path_to_config_file(server['ssh_key'])} \
        -p {server['ssh_port']} \
        -o StrictHostKeyChecking=no \
        {run}"
    while command.find("  ") != -1:
        command = command.replace("  ", " ")
    print(command)
    subprocess.check_call(command.split(" "))


def ssh(run="", index=None, as_root=False):
    set_config(sys.argv[1])

    if index is not None:
        server = get_config()["servers"][index]
        ssh_server(server, run, as_root)
    else:
        for server in get_config()["servers"]:
            ssh_server(server, run, as_root)


if __name__ == "__main__":
    index = int(sys.argv[2])
    ssh(index=index)



import subprocess
import sys

from ..helpers.config import get_config, path_to_config_file, set_config


def sftp(index):
    set_config(sys.argv[1])

    server = get_config()["servers"][index]
    command = f"sftp \
        -i {path_to_config_file(server['ssh_key'])} \
        -P {server['ssh_port']} \
        -o StrictHostKeyChecking=no \
        {server['ssh_user']}@{server['ssh_hostname']}"
    while command.find("  ") != -1:
        command = command.replace("  ", " ")
    print(command)
    subprocess.check_call(command.split(" "))


if __name__ == "__main__":
    index = int(sys.argv[2])
    sftp(index=index)



from ...helpers.fs import make_directory, run_with_marker
from ...helpers.network import wireguard_ip

options = {}


def dependencies():
    from . import init

    return [(init, options)]


def make_user(username: str, password: str):
    from .common import CERTS_DIR, HOME_DIR, cockroach_binary

    user_dir = HOME_DIR.joinpath("users")
    make_directory(user_dir)
    SQL_PORT = options.get("SQL_PORT", 26258)
    run_with_marker(
        user_dir.joinpath(username),
        f"{cockroach_binary} sql --certs-dir={CERTS_DIR} --host={wireguard_ip()}:{SQL_PORT} --execute \"CREATE USER {username} WITH PASSWORD '{password}';\"",
        run_if_command_changed=False,
    )


def make_db(name: str, owner: str):
    from .common import CERTS_DIR, HOME_DIR, cockroach_binary

    data_dir = HOME_DIR.joinpath("databases")
    make_directory(data_dir)
    SQL_PORT = options.get("SQL_PORT", 26258)
    run_with_marker(
        data_dir.joinpath(name),
        f'{cockroach_binary} sql --certs-dir={CERTS_DIR} --host={wireguard_ip()}:{SQL_PORT} --execute="CREATE DATABASE {name} OWNER {owner};"',
        run_if_command_changed=False,
    )



import os
import re
from glob import glob
from pathlib import Path
from typing import Dict, List, Optional, Union

from debian.debian_support import version_compare

from paracrine import dry_run_safe_read

from .fs import (
    build_with_command,
    download,
    run_command,
    run_with_marker,
    set_file_contents,
)

host_arch: Optional[str] = None

_version_pattern = re.compile(r"Version: (\S+)")


def apt_update():
    run_with_marker(
        "/opt/apt-update",
        "apt-get update --allow-releaseinfo-change",
        deps=glob("/etc/apt/sources.list.d/*")
        + glob("/etc/apt/trusted.gpg.d/*")
        + ["/etc/apt/sources.list"],
    )


def add_trusted_key(url: str, name: str, hash: str, armored: bool = True):
    if armored:
        download(
            url,
            f"/etc/apt/trusted.gpg.d/{name}.gpg.asc",
            hash,
        )
        apt_install(["gpg"])
        build_with_command(
            f"/etc/apt/trusted.gpg.d/{name}.gpg",
            f"cat /etc/apt/trusted.gpg.d/{name}.gpg.asc | gpg --dearmor > /etc/apt/trusted.gpg.d/{name}.gpg",
            [f"/etc/apt/trusted.gpg.d/{name}.gpg.asc"],
        )
    else:
        download(
            url,
            f"/etc/apt/trusted.gpg.d/{name}.gpg",
            hash,
        )
    return f"/etc/apt/trusted.gpg.d/{name}.gpg"


def apt_is_installed(package: str, wanted_version: Optional[str] = None) -> bool:
    paths = [
        f"/var/lib/dpkg/info/{package}.list",
        f"/var/lib/dpkg/info/{package}:{host_arch}.list",
    ]
    for path in paths:
        if not os.path.exists(path):
            continue
        if wanted_version is None:  # existance is enough
            return True
        status = run_command(f"dpkg-query --status {package}", dry_run_safe=True)
        version_pattern_match = _version_pattern.search(status)
        if version_pattern_match is None:
            raise Exception(
                f"Failure to match version pattern in '{status}' for {package}"
            )
        version = version_pattern_match.group(1)
        if version_compare(version, wanted_version) >= 0:
            return True

    return False


# List is just "any version", Dict is a "name => min version" requirement
def apt_install(
    packages: Union[List[str], Dict[str, Optional[str]]],
    always_install: bool = False,
    target_release: Optional[str] = None,
) -> bool:
    global host_arch
    if host_arch is None and packages != ["dpkg-dev"]:
        apt_install(["dpkg-dev"])
        host_arch_path = Path("/opt/host-arch")
        build_with_command(
            host_arch_path, f"dpkg-architecture -q DEB_HOST_ARCH > {host_arch_path}"
        )
        host_arch = dry_run_safe_read(host_arch_path, "amd64")
    if isinstance(packages, List):
        packages = dict([(p, None) for p in packages])
    if always_install:
        to_install = packages
    else:
        to_install = {}
        for package in packages.keys():
            wanted_version = packages[package]
            if not apt_is_installed(package, wanted_version):
                to_install[package] = wanted_version

        if to_install == {}:
            return False

    apt_update()
    # Confdef is to fix https://unix.stackexchange.com/a/416816/73838
    os.environ["DEBIAN_FRONTEND"] = "noninteractive"
    cmd = (
        "apt-get install %s --no-install-recommends --yes -o DPkg::Options::=--force-confdef"
        % " ".join(
            [
                name if version is None else f"{name}>={version}"
                for (name, version) in to_install.items()
            ]
        )
    )
    if target_release is not None:
        cmd += f" --target-release {target_release}"
    run_command(cmd)
    return True


def debian_repo(name, contents=None):
    fname = "/etc/apt/sources.list.d/%s.list" % name
    if contents is None:
        contents = "deb http://deb.debian.org/debian %s main" % name

    changed = set_file_contents(fname, contents)
    if changed:
        apt_update()
    return changed


def set_alternative(alt_name: str, option: str):
    current_alt_path = os.path.realpath(f"/etc/alternatives/{alt_name}")
    if current_alt_path != option:
        print(
            f"current alt path for {alt_name} is {current_alt_path} not {option}, so fixing"
        )
        run_command(f"update-alternatives --set {alt_name} {option}")
        return True
    else:
        return False



from typing import Dict, List

from ...helpers.config import other_config_file
from ...helpers.fs import delete
from .common import MASTER_FILE


def run():
    pass


def parse_return(infos: List[Dict]) -> None:
    delete(other_config_file(MASTER_FILE), True)



import contextlib
import grp
import hashlib
import logging
import os
import pwd
import re
import stat
import subprocess
from datetime import datetime, timedelta
from difflib import unified_diff
from pathlib import Path
from typing import List, Optional, Sequence, Union

from typing_extensions import TypedDict

from paracrine import is_dry_run

from .config import data_files, jinja_env

Pathy = Union[str, Path]


def hash_data(data: bytes) -> str:
    m = hashlib.sha256()
    m.update(data)
    return m.hexdigest()


def set_file_contents(
    fname: Pathy,
    contents: Union[str, bytes],
    ignore_changes: bool = False,
    owner: Optional[str] = None,
    group: Optional[str] = None,
) -> bool:
    needs_update = False

    if type(contents) == bytes:
        try:
            contents = contents.decode("utf-8")
        except UnicodeDecodeError:
            pass

    if not os.path.exists(fname):
        needs_update = True
        logging.info("File %s was missing" % fname)
    elif not ignore_changes:
        if isinstance(contents, str):
            data = open(fname, "rb").read().decode("utf-8").splitlines(True)
            diff = list(unified_diff(data, contents.splitlines(True)))
            if len(diff) > 0:
                diff = "".join(diff)
                logging.info("File %s was different. Diff is: \n%s" % (fname, diff))
                needs_update = True
        else:
            data = open(fname, "rb").read()
            if hash_data(data) != hash_data(contents):
                logging.info("File %s was different" % fname)
                needs_update = True

    if needs_update and not is_dry_run():
        if isinstance(contents, str):
            open(fname, "w").write(contents)
        else:
            open(fname, "wb").write(contents)

    needs_update = set_owner(fname, owner, group) or needs_update

    return needs_update


def render_template(template: str, **kwargs: object) -> str:
    return jinja_env().get_template(template).render(**kwargs)


def set_file_contents_from_template(
    fname: Pathy, template: str, ignore_changes: bool = False, **kwargs: object
) -> bool:
    return set_file_contents(
        fname,
        render_template(template, **kwargs),
        ignore_changes=ignore_changes,
    )


def set_file_contents_from_data(fname: str, data_path: str):
    return set_file_contents(fname, data_files()[data_path])


@contextlib.contextmanager
def cd(path: Pathy):
    CWD = os.getcwd()

    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(CWD)


def set_mode(path: Pathy, mode: Union[str, int]) -> bool:
    if isinstance(mode, str):
        raw_mode = int(mode, 8)
    else:
        raw_mode = mode

    dry_run = is_dry_run()

    try:
        existing = stat.S_IMODE(os.stat(path).st_mode)
    except FileNotFoundError:
        if dry_run:
            logging.info(f"Missing {path}, but would have set mode to {mode}")
            return True
        else:
            raise
    if existing != raw_mode:
        logging.info("chmod %s %s" % (path, mode))
        if not dry_run:
            os.chmod(path, raw_mode)
        return True
    else:
        return False


def set_owner(
    path: Pathy, owner: Optional[str] = None, group: Optional[str] = None
) -> bool:
    if owner is None and group is None:
        return False
    try:
        st = os.stat(path)
    except FileNotFoundError:
        if is_dry_run():
            logging.info(
                f"Can't find {path}, but would have set owner {owner} and group {group}"
            )
            return True
        else:
            raise
    if owner is not None:
        owner_id = pwd.getpwnam(owner).pw_uid
    else:
        owner_id = st.st_uid
    if group is not None:
        group_id = grp.getgrnam(group).gr_gid
    else:
        group_id = st.st_gid

    if st.st_uid != owner_id or st.st_gid != group_id:
        if not is_dry_run():
            os.chown(path, owner_id, group_id)
        return True
    else:
        return False


def make_directory(
    path: Pathy,
    mode: Union[str, int, None] = None,
    owner: Optional[str] = None,
    group: Optional[str] = None,
) -> bool:
    ret = False
    if not os.path.exists(path):
        logging.info("Make directory %s" % path)
        if not is_dry_run():
            os.makedirs(path)
        ret = True
    if mode is not None:
        ret = set_mode(path, mode) or ret
    if owner is not None or group is not None:
        set_owner(path, owner, group)
    return ret


def replace_line(fname: Pathy, search: str, replace: str) -> bool:
    existing = open(fname).read()
    if search in existing:
        return set_file_contents(fname, existing.replace(search, replace))
    else:
        return False


def insert_line(fname: Pathy, line: str) -> bool:
    existing = open(fname).read()
    if line not in existing:
        return set_file_contents(fname, existing + "\n" + line)
    else:
        return False


def insert_or_replace(fname: str, matcher: Union[re.Pattern, str], line: str) -> bool:
    existing = open(fname).read()
    if isinstance(matcher, re.Pattern):
        results = matcher.search(existing)
        if results is not None:
            return set_file_contents(
                fname, existing[: results.start()] + line + existing[results.end() :]
            )
    elif matcher in existing:
        return set_file_contents(fname, existing.replace(matcher, line))
    return set_file_contents(fname, existing + "\n" + line)


def sha_file(fname: Pathy) -> str:
    from .debian import apt_install

    apt_install(["coreutils"])
    existing_sha = run_command("sha256sum %s" % fname, dry_run_safe=True).strip()
    return existing_sha.split(" ")[0]


def has_sha(fname: Pathy, sha: str) -> bool:
    if os.path.exists(fname):
        existing_sha = sha_file(fname)
        if existing_sha == sha:
            return True

    return False


def download(
    url: str, fname: Pathy, sha: str, mode: Union[int, str, None] = None
) -> bool:
    exists = has_sha(fname, sha)
    if not exists:
        from .debian import apt_install

        if not is_dry_run():
            apt_install(["curl", "ca-certificates"])
            run_command("curl -Lo %s %s" % (fname, url))
            existing_sha = sha_file(fname)
            assert existing_sha == sha, (existing_sha, sha)

    if mode is not None:
        set_mode(fname, mode)

    return not exists


def link(target: Pathy, source: Pathy) -> bool:
    if os.path.lexists(target) and (
        not os.path.exists(target) or not os.path.samefile(source, target)
    ):
        logging.info("Unlink %s" % target)
        if not is_dry_run():
            os.remove(target)
    if not os.path.lexists(target):
        logging.info("Link %s to %s" % (target, source))
        if not is_dry_run():
            os.symlink(source, target)
        return True
    else:
        return False


def download_executable(
    url: str, hash: str, name: Optional[str] = None, path: Optional[Pathy] = None
) -> bool:
    if name is None:
        name = url.split("/")[-1]
    if path is None:
        path = "/usr/local/bin/%s" % name
    return download(
        url,
        path,
        hash,
        mode="755",
    )


class Unpacked(TypedDict):
    changed: bool
    dir_name: Pathy


def download_and_unpack(
    url: str,
    hash: str,
    name: Optional[str] = None,
    dir_name: Optional[Pathy] = None,
    compressed_root: str = "/opt",
) -> Unpacked:
    if name is None:
        name = url.split("/")[-1]
    compressed_path: str = "%s/%s" % (compressed_root, name)
    if dir_name is None:
        dir_name = "/opt/%s" % name.replace(".tar.gz", "").replace(".tgz", "").replace(
            ".tar.xz", ""
        )
    changed = download(
        url,
        compressed_path,
        hash,
    )

    make_directory(dir_name)
    marker_name = Path(compressed_path + ".unpacked")
    if not marker_name.exists():
        from .debian import apt_install

        if compressed_path.endswith("tar.gz") or compressed_path.endswith(".tgz"):
            apt_install(["tar"])
            run_command("tar --directory=%s -zxvf %s" % (dir_name, compressed_path))
        elif compressed_path.endswith("tar.xz"):
            apt_install(["tar", "xz-utils"])
            run_command("tar --directory=%s -Jxvf %s" % (dir_name, compressed_path))
        elif compressed_path.endswith("zip"):
            apt_install(["unzip"])
            run_command("unzip %s -d %s" % (compressed_path, dir_name))
        else:
            raise Exception(compressed_path)

        set_file_contents(marker_name, "")

        changed = True

    return {"changed": changed, "dir_name": dir_name}


def last_modified(fname: Pathy) -> float:
    try:
        return os.stat(fname).st_mtime
    except FileNotFoundError:
        return float(0)


def delete(fname: Pathy, quiet: bool = False) -> bool:
    if os.path.exists(fname):
        if not quiet:
            logging.info("Deleting %s", fname)
        if not is_dry_run():
            os.remove(fname)
        return True
    else:
        return False


def build_with_command(
    fname: Pathy,
    command: str,
    deps: List[Pathy] = [],
    force_build: bool = False,
    directory: Optional[Pathy] = None,
) -> bool:
    display = command.strip()
    while display.find("  ") != -1:
        display = display.replace("  ", " ")
    changed = set_file_contents("%s.command" % fname, display) or force_build
    target_modified = last_modified(fname)
    for dep in deps:
        if last_modified(dep) > target_modified:
            logging.info("%s is younger than %s" % (dep, fname))
            changed = True
            break
    if not os.path.exists(fname) or changed:
        logging.info("Building %s" % fname)
        if command.find("|") != -1:
            out = ""
            commands = command.split("|")
            for command in commands:
                try:
                    out = run_command(command, input=out, directory=directory)
                except subprocess.CalledProcessError:
                    print("Ran '%s' with input '%s'" % (command, out))
                    raise
        else:
            run_command(command, directory=directory)
        return True
    else:
        return False


def run_with_marker(
    fname: Pathy,
    command: str,
    deps: Sequence[Pathy] = [],
    max_age: Optional[timedelta] = None,
    force_build: bool = False,
    directory: Optional[Pathy] = None,
    run_if_command_changed: bool = True,
    input: Optional[str] = None,
    dry_run_safe: bool = False,
) -> bool:
    changed = not os.path.exists(fname) or force_build
    target_modified = last_modified(fname)
    if max_age is not None:
        age = datetime.now() - datetime.fromtimestamp(target_modified)
        if age > max_age:
            changed = True
    for dep in deps:
        dep_modified = last_modified(dep)
        if dep_modified > target_modified:
            logging.info("%s is younger than %s" % (dep, fname))
            changed = True
            break

    if run_if_command_changed and not changed:
        old_command = open(fname).read()
        changed = old_command != command

    if changed:
        run_command(
            command, directory=directory, input=input, dry_run_safe=dry_run_safe
        )
        if dry_run_safe or not is_dry_run():
            open(fname, "w").write(command)

    return changed


class MissingCommandException(Exception):
    pass


def run_command(
    cmd: str,
    directory: Optional[Pathy] = None,
    input: Optional[str] = None,
    allowed_exit_codes: List[int] = [0],
    dry_run_safe: bool = False,
) -> str:
    run_for_real = dry_run_safe or not is_dry_run()
    display = cmd.strip()
    while display.find("  ") != -1:
        display = display.replace("  ", " ")
    try:
        process = None
        if directory is not None:
            if run_for_real:
                logging.info("Run in %s: %s" % (directory, display))
                with cd(directory):
                    process = subprocess.Popen(
                        cmd,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,
                        stdin=subprocess.PIPE,
                        encoding="utf-8",
                    )
            else:
                logging.info("Would have run in %s: %s" % (directory, display))
        else:
            if run_for_real:
                logging.info("Run: %s" % display)
                process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    stdin=subprocess.PIPE,
                    encoding="utf-8",
                )
            else:
                logging.info("Would have run: %s" % display)

        if run_for_real:
            assert process is not None
            (stdout, stderr) = process.communicate(input=input)
            if process.returncode not in allowed_exit_codes:
                if ": not found" in stderr:
                    # missing command
                    raise MissingCommandException
                assert process.returncode in allowed_exit_codes, (
                    process.returncode,
                    stdout,
                    stderr,
                )
            return stdout
        else:
            return ""
    except subprocess.CalledProcessError as e:
        print(e.output)
        raise



import json
from typing import Dict, List

from ...helpers.config import other_config_file
from ...helpers.fs import set_file_contents
from .. import wireguard
from . import wipe_old_master
from .common import MASTER_FILE, get_existing_masters, get_master_ip

options = {}


def dependencies():
    return [wireguard, wipe_old_master]


def run():
    master_ip = get_master_ip()
    return {"master_ip": master_ip}


def parse_return(infos: List[Dict]) -> None:
    existing_masters = get_existing_masters(True)
    existing_masters.append(infos[0]["master_ip"])
    master_path = other_config_file(MASTER_FILE)
    set_file_contents(master_path, json.dumps(existing_masters, indent=2))



import json
import os
import socket
from pathlib import Path
from typing import Dict, List

from paracrine import is_dry_run

from ..helpers.config import (
    ServerDict,
    add_return_data,
    config,
    config_path,
    host,
    in_docker,
    network_config_file,
    other_config,
    other_config_file,
)
from ..helpers.debian import apt_install
from ..helpers.fs import (
    MissingCommandException,
    make_directory,
    run_command,
    set_file_contents,
)
from ..helpers.users import in_vagrant, users


def is_wireguard():
    return os.path.exists("/etc/wireguard")


def hash_fn(key: str, count: int) -> int:
    return sum(bytearray(key.encode("utf-8"))) % count


def _index_fn(name: str) -> ServerDict:
    hosts = config()["servers"]
    try:
        existing_selectors = other_config("selectors.json")
        return [host for host in hosts if host["name"] == existing_selectors[name]][0]
    except KeyError:
        index = hash_fn(name, len(hosts))
        add_return_data({"selector": {name: hosts[index]["name"]}})
        return hosts[index]


# Use this host for a given service
# Intended for "run on one machine" things
def use_this_host(name: str) -> bool:
    use_host = _index_fn(name)
    return host()["name"] == use_host["name"]


def wireguard_ip_for_machine_for(name: str) -> str:
    use_host = _index_fn(name)
    return use_host.get("wireguard_ip", "<unknown>")


def run():
    apt_install(["iproute2"])

    data = {
        "hostname": socket.gethostname(),
        "users": users(force_load=True),
        "groups": run_command("getent group", dry_run_safe=True),
        "server_name": host()["name"],
    }
    try:
        data["network_devices"] = run_command("ip -j address", dry_run_safe=True)
    except MissingCommandException:
        if is_dry_run():
            data["network_devices"] = "{}"
        else:
            raise
    ip_file = Path("/opt/ip_address")
    if ip_file.exists():
        data["external_ip"] = json.load(ip_file.open())
    else:
        if in_vagrant() or in_docker():
            networks = json.loads(data["network_devices"])
            ext_if = [
                net
                for net in networks
                if net["ifname"].startswith("eth") and len(net["addr_info"]) > 0
            ]
            if len(ext_if) > 0:
                data["external_ip"] = json.dumps(
                    {"ip": ext_if[-1]["addr_info"][0]["local"]}
                )
            else:
                data["external_ip"] = "<unknown>"
        else:
            apt_install(["curl", "ca-certificates"])
            data["external_ip"] = run_command(
                "curl https://api.ipify.org?format=json", dry_run_safe=True
            )
        set_file_contents(ip_file, json.dumps(data["external_ip"]))

    return data


def parse_return(infos: List[Dict]) -> None:
    info = infos[0]
    networks = json.loads(info["network_devices"])
    name = info["server_name"]
    make_directory(config_path())
    set_file_contents(network_config_file(name), json.dumps(networks, indent=2))

    other = {
        "users": info["users"],
        "groups": info["groups"],
        "hostname": info["hostname"],
    }
    try:
        other["external_ip"] = json.loads(info["external_ip"])["ip"]
    except json.JSONDecodeError:
        if is_dry_run():
            other["external_ip"] = "<unknown>"
        else:
            raise

    set_file_contents(other_config_file(name), json.dumps(other, indent=2))



from .config import other_self_config
from .fs import run_command


def users(force_load=False):
    if not force_load:
        try:
            return other_self_config()["users"]
        except KeyError:
            pass

    raw_users = run_command("getent passwd | cut -d: -f1", dry_run_safe=True)
    return sorted(raw_users.strip().split("\n"))


def adduser(name, home_dir=None):
    if name not in users():
        extra = ""
        if home_dir is not None:
            extra += f"--home-dir {home_dir} --create-home"
        run_command(f"useradd {name} {extra}")
        return True
    else:
        return False


def groups():
    raw_groups = other_self_config()["groups"]

    ret = {}
    for line in raw_groups.split("\n"):
        if line == "":
            continue
        bits = line.split(":")
        if len(bits) < 3:
            raise Exception((line, bits))
        ret[bits[0]] = bits[3].split(",")

    return ret


def add_user_to_group(user, group):
    existing_groups = groups()
    existing_group = existing_groups[group]
    if user not in existing_group:
        run_command("usermod -aG %s %s" % (group, user))
        return True
    else:
        return False


def in_vagrant():
    return "vagrant" in users()



import json
import pathlib
from typing import Optional

from paracrine import is_dry_run

from .config import config_path, get_config_file
from .fs import delete, set_file_contents, set_file_contents_from_template

mailto: Optional[str] = None
mailfrom: Optional[str] = None


def local():
    global mailto, mailfrom
    cron_path = pathlib.Path(config_path()).joinpath("cron-info")
    set_file_contents(
        cron_path,
        json.dumps(
            {"mailto": mailto, "mailfrom": mailfrom},
            indent=2,
        ),
    )


def set_mailto(email: str) -> None:
    global mailto
    mailto = email


def set_mailfrom(email: str) -> None:
    global mailfrom
    mailfrom = email


def cron_path(name: str):
    return f"/etc/cron.d/{name}"


def create_cron(name: str, schedule: str, user: str, command: str):
    try:
        cron_info = json.loads(get_config_file("configs/cron-info"))
    except KeyError:
        if not is_dry_run():
            raise
        cron_info = {}
    envs = {}
    if cron_info.get("mailto") is not None:
        envs["MAILTO"] = cron_info["mailto"]
    if cron_info.get("mailfrom") is not None:
        envs["MAILFROM"] = cron_info["mailfrom"]
    set_file_contents_from_template(
        cron_path(name),
        "cron.j2",
        SCHEDULE=schedule,
        USER=user,
        COMMAND=command,
        ENVS=envs,
    )


def delete_cron(name: str):
    delete(cron_path(name))



from pathlib import Path

from ..helpers.debian import add_trusted_key, apt_install, debian_repo
from ..helpers.fs import make_directory, run_with_marker
from ..helpers.systemd import systemd_set


def run():
    add_trusted_key(
        "https://www.postgresql.org/media/keys/ACCC4CF8.asc",
        "postgresql",
        "0144068502a1eddd2a0280ede10ef607d1ec592ce819940991203941564e8e76",
    )
    debian_repo(
        "postgresql_org_repository",
        "deb http://apt.postgresql.org/pub/repos/apt bullseye-pgdg main",
    )
    apt_install(["postgresql-14"], target_release="bullseye-pgdg")
    systemd_set("postgresql@14-main", enabled=True, running=True)


db_dir = Path("/opt/postgresql")


def make_user(username: str, password: str):
    user_dir = db_dir.joinpath("users")
    make_directory(user_dir)
    run_with_marker(
        user_dir.joinpath(username),
        f"sudo -u postgres psql --command=\"CREATE USER {username} WITH ENCRYPTED PASSWORD '{password}';\"",
        run_if_command_changed=False,
    )


def make_db(name: str, owner: str):
    data_dir = db_dir.joinpath("databases")
    make_directory(data_dir)
    run_with_marker(
        data_dir.joinpath(name),
        f'sudo -u postgres psql --command="CREATE DATABASE {name} OWNER {owner};"',
        run_if_command_changed=False,
    )



import sys
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Union, cast

from paracrine import dry_run_safe_read

from ..deps import Modules
from ..helpers import cron
from ..helpers.config import core_config, local_config, other_config_file
from ..helpers.debian import apt_install
from ..helpers.fs import (
    make_directory,
    run_command,
    run_with_marker,
    set_file_contents,
    set_file_contents_from_template,
)
from ..helpers.python import setup_venv
from . import aws
from .core import use_this_host

options = {}


# Are we in a test config where we should just not get the cert
def get_dummy_certs() -> bool:
    try:
        config = local_config()
    except FileNotFoundError:
        config = core_config()
    return cast(bool, config.get("dummy_certs", False))


def certbot_for_host(hostname: Union[str, List[str]], email: str) -> Dict[str, object]:
    if isinstance(hostname, str):
        hostnames = [hostname]
    else:
        hostnames = hostname
    certbot = Path("/opt/certbot")
    cert_name = "_".join(hostnames)
    live_path = certbot.joinpath("config", "live", cert_name)

    dummy_certs = get_dummy_certs()

    if use_this_host("certbot"):
        venv = certbot.joinpath("venv")
        venv_bin = venv.joinpath("bin")
        pip = venv_bin.joinpath("pip")
        certbot_bin = venv_bin.joinpath("certbot")

        fullchain_path = live_path.joinpath("fullchain.pem")
        envs = aws.get_env_with_creds()
        renew_command = f"{envs} {certbot_bin} renew \
                        --cert-name={cert_name} \
                        --config-dir={certbot.joinpath('config')} \
                        --work-dir={certbot.joinpath('workdir')} \
                        --logs-dir={certbot.joinpath('logs')} \
                        --dns-route53 \
                        --no-random-sleep-on-renew"
        while renew_command.find("  ") != -1:
            renew_command = renew_command.replace("  ", " ")
        make_directory(live_path)
        make_directory(certbot)
        setup_venv(venv)
        set_file_contents_from_template(
            "/opt/certbot/requirements.txt", "certbot_requirements.txt"
        )
        run_with_marker(
            "/opt/certbot/deps_installed",
            f"{pip} install -r /opt/certbot/requirements.txt",
            deps=["/opt/certbot/requirements.txt"],
        )
        config_path = certbot.joinpath(f"config/renewal/{cert_name}.conf")

        if not config_path.exists():
            if dummy_certs:
                set_file_contents(fullchain_path, "")
                set_file_contents(live_path.joinpath("privkey.pem"), "")
            else:
                run_command(
                    f"{envs} {certbot_bin} certonly \
                        --config-dir={certbot.joinpath('config')} \
                        --work-dir={certbot.joinpath('workdir')} \
                        --logs-dir={certbot.joinpath('logs')} \
                        --cert-name={cert_name} \
                        -m {email} --agree-tos --non-interactive \
                        --no-eff-email --domains {','.join(hostnames)} --dns-route53"
                )
        else:
            if not dummy_certs:
                run_with_marker(
                    f"/opt/certbot/renew_marker_{cert_name}",
                    renew_command,
                    max_age=timedelta(days=1),
                )

        apt_install(["moreutils"])
        cron.create_cron(
            f"certs-renew-{hostnames[0]}",
            "0 3 * * *",
            "root",
            f"chronic {renew_command}",
        )

        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

        return {
            f"fullchain-{hostnames[0]}": dry_run_safe_read(
                fullchain_path, "dummy fullchain"
            ),
            f"privkey-{hostnames[0]}": dry_run_safe_read(
                live_path.joinpath("privkey.pem"), "dummy privkey"
            ),
            "ssl-options": dry_run_safe_read(
                venv.joinpath(
                    f"lib/python{python_version}/site-packages/certbot_nginx/_internal/tls_configs/options-ssl-nginx.conf"  # noqa: E501
                ),
                "fake ssl options",
            ),
        }
    else:
        return {}


def dependencies() -> Modules:
    return [aws, cron]


def run() -> Dict:
    return certbot_for_host(options["hostname"], options["email"])


def parse_return(
    infos: List[Dict],
) -> None:
    for info in infos:
        for key in info:
            if key == "selector":
                continue
            open(other_config_file(key), "w").write(info[key])



import tempfile
from typing import BinaryIO, cast

import pytest
import yaml

from paracrine.runner import run


def set_config_data(config_file: BinaryIO, data: object) -> None:
    config_file.write(yaml.dump(data).encode("utf-8"))
    config_file.flush()


def test_empty_servers():
    with tempfile.NamedTemporaryFile() as config_file:
        set_config_data(cast(BinaryIO, config_file), {"servers": []})
        run(["-i", config_file.name], [])


def test_bad_ssh_path():
    with tempfile.NamedTemporaryFile() as config_file:
        set_config_data(
            cast(BinaryIO, config_file),
            {
                "data_path": ".",
                "servers": [{"ssh_hostname": "", "ssh_key": "TBD", "ssh_user": ""}],
            },
        )
        with pytest.raises(Exception, match="Can't find ssh key /tmp/configs/TBD"):
            run(["-i", config_file.name], [])



from ...helpers.config import build_config, core_config
from ...helpers.fs import run_command
from ...helpers.network import wireguard_ip, wireguard_ips
from . import check_master, node
from .common import get_master_ip

options = {}


def dependencies():
    return [(node, options), check_master]


def run():
    master_ip = get_master_ip()
    local_ip = wireguard_ip()

    LOCAL = build_config(core_config())

    output = run_command(
        f"redis-cli -a {LOCAL['REDIS_PASSWORD']} -h {local_ip} info replication"
    )
    if "role:slave" in output:
        assert f"master_host:{master_ip}" in output, output
        assert "master_link_status:up" in output, output
    else:
        assert "role:master" in output, output
        for ip in wireguard_ips().values():
            if ip == master_ip:
                continue
            assert f"ip={ip},port=6379,state=online" in output, output

    output = run_command(
        f"redis-cli -a {LOCAL['REDIS_PASSWORD']} -h {local_ip} -p 26379 info sentinel"
    )
    count = len(wireguard_ips())
    assert (
        f"master0:name=mymaster,status=ok,address={master_ip}:6379,slaves={count-1},sentinels={count}"
        in output
    ), output



from pathlib import Path
from typing import Dict, List, Optional

from paracrine import dry_run_safe_read

from ...helpers.config import config_path
from ...helpers.fs import (
    build_with_command,
    download_and_unpack,
    make_directory,
    set_file_contents,
)
from ...helpers.network import wireguard_ips
from ...runners.core import use_this_host
from .common import binary_path, cockroach_hash, cockroach_url, local_node_ip

options = {}


def run():
    if not use_this_host("cockroach-certs"):
        return None
    unpacked = download_and_unpack(
        cockroach_url,
        cockroach_hash,
    )
    cockroach = Path(unpacked["dir_name"]).joinpath(binary_path)
    certs_dir = Path("/opt/cockroach/certs")
    make_directory(certs_dir)
    ca_key_path = certs_dir.joinpath("ca.key")
    build_with_command(
        ca_key_path,
        f"{cockroach} cert create-ca --certs-dir={certs_dir} --ca-key={ca_key_path} --overwrite --allow-ca-key-reuse",
    )
    root_key_path = certs_dir.joinpath("client.root.key")
    build_with_command(
        root_key_path,
        f"{cockroach} cert create-client root --certs-dir={certs_dir} --ca-key={ca_key_path} --overwrite",
        deps=[ca_key_path],
    )

    node_keys = {}
    ip_keys = set(wireguard_ips().values())
    ip_keys.add(local_node_ip())
    for ip in ip_keys:
        crt_path = certs_dir.joinpath(f"{ip}.crt")
        key_path = crt_path.with_suffix(".key")
        build_with_command(
            crt_path,
            f"{cockroach} cert create-node localhost {ip} --certs-dir={certs_dir} --ca-key={ca_key_path} --overwrite && mv {certs_dir.joinpath('node.crt')} {crt_path} && mv {certs_dir.joinpath('node.key')} {key_path}",
            deps=[ca_key_path],
        )
        node_keys[ip] = {
            "crt": dry_run_safe_read(crt_path, "fake crt"),
            "key": dry_run_safe_read(key_path, "fake key"),
        }

    return {
        "ca_crt": dry_run_safe_read(ca_key_path.with_suffix(".crt"), "fake crt"),
        "root_key": dry_run_safe_read(root_key_path, "fake root key"),
        "root_crt": dry_run_safe_read(
            root_key_path.with_suffix(".crt"), "fake root crt"
        ),
        "node_keys": node_keys,
    }


def parse_return(infos: List[Optional[Dict]]):
    assert len(infos) == 1, infos
    info = infos[0]
    if info is None or "ca_crt" not in info:
        return
    certs_dir = Path(config_path()).joinpath("cockroach-certs")
    make_directory(certs_dir)
    set_file_contents(certs_dir.joinpath("ca.crt"), info["ca_crt"])
    set_file_contents(certs_dir.joinpath("client.root.key"), info["root_key"])
    set_file_contents(certs_dir.joinpath("client.root.crt"), info["root_crt"])

    for ip, values in info["node_keys"].items():
        set_file_contents(certs_dir.joinpath(f"{ip}.key"), values["key"])
        set_file_contents(certs_dir.joinpath(f"{ip}.crt"), values["crt"])



from pathlib import Path

from ...helpers.config import in_docker
from ...helpers.network import wireguard_ip

cockroach_version = "23.1.1"
cockroach_url = (
    f"https://binaries.cockroachdb.com/cockroach-v{cockroach_version}.linux-amd64.tgz"
)
cockroach_hash = "8197562ce59d1ac4f53f67c9d277827d382db13c5e650980942bcb5e5104bb4e"
binary_path = f"cockroach-v{cockroach_version}.linux-amd64/cockroach"
# FIXME
cockroach_binary = f"/opt/cockroach-v{cockroach_version}.linux-amd64/{binary_path}"
HOME_DIR = Path("/var/lib/cockroach")
CERTS_DIR = HOME_DIR.joinpath("certs")
USER = "cockroach"


def local_node_ip():
    if in_docker():
        return "127.0.0.1"
    else:
        return wireguard_ip()



import json
import os
import sys
from distutils.version import LooseVersion
from pathlib import Path
from typing import Dict, List

from typing_extensions import TypedDict

from paracrine import dry_run_safe_read, is_dry_run

from ...helpers.config import (
    get_config,
    host,
    in_docker,
    network_config,
    other_config_file,
)
from ...helpers.debian import apt_install
from ...helpers.fs import (
    MissingCommandException,
    make_directory,
    run_command,
    set_file_contents,
)
from .common import private_key_file, public_key_file, public_key_path, wg_config


def get_all_kernel_versions():
    raw = run_command(
        r"dpkg-query --showformat=$\{Package\},$\{Status\},$\{Version\}\\t --show linux-image-*",
    )
    versions: Dict[str, str] = {}
    for line in raw.split("\t"):
        if line.strip() == "":
            continue
        try:
            (name, status, version) = line.split(",")
        except ValueError:
            raise Exception("'%s'" % line)
        if status == "install ok installed":
            versions[name] = version
    return versions


class ReturnDict(TypedDict):
    wg_publickey: str
    host: str


def run() -> ReturnDict:
    apt_install(["kmod", "wireguard"])
    try:
        modules = sorted(
            [
                line.split(" ")[0]
                for line in run_command("lsmod", dry_run_safe=True).splitlines()
            ]
        )
    except MissingCommandException:
        if is_dry_run():
            modules = []
        else:
            raise
    if "wireguard" not in modules and not in_docker():
        print("modules", modules)
        apt_install(["linux-image-amd64"])
        versions = get_all_kernel_versions()
        version_keys = [
            x.replace("linux-image-", "")
            for x in versions.keys()
            if x
            not in [
                "linux-image-amd64",
                "linux-image-cloud-amd64",
                "linux-image-rt-amd64",
            ]
        ]
        print("kernel version keys", version_keys)
        ordered = sorted(
            version_keys,
            key=LooseVersion,
            reverse=True,
        )
        highest = ordered[0]
        current = run_command("uname -r").strip()
        if current != highest:
            print(ordered)
            print("'%s' != '%s'" % (current, highest))
            apt_install(["systemd-sysv"])
            if not in_docker():
                run_command("reboot")
                sys.exit(0)

        if not in_docker():
            apt_install(["linux-headers-amd64"])
            try:
                modules = sorted(
                    [
                        line.split(" ")[0]
                        for line in run_command("lsmod", dry_run_safe=True).splitlines()
                    ]
                )
            except MissingCommandException:
                if is_dry_run():
                    modules = []
                else:
                    raise
            if "wireguard" not in modules:
                print("modules", modules)
                run_command("modprobe wireguard")

    make_directory(wg_config)
    if not os.path.exists(private_key_file):
        run_command("wg genkey > %s" % private_key_file)
    if not os.path.exists(public_key_file):
        run_command("cat %s | wg pubkey > %s" % (private_key_file, public_key_file))

    return {
        "wg_publickey": dry_run_safe_read(Path(public_key_file), "dummy wg publickey"),
        "host": host()["name"],
    }


def parse_return(infos: List[ReturnDict]) -> None:
    assert len(infos) == 1, infos
    info = infos[0]
    set_file_contents(public_key_path(info["host"]), info["wg_publickey"])

    wg_ips = []

    for server in get_config()["servers"]:
        networks = network_config(server["name"])
        wireguard_networks = [
            network for network in networks if network["ifname"] == "wg0"
        ]
        if len(wireguard_networks) == 1:
            wg_ips.append(wireguard_networks[0]["addr_info"][0]["local"])
    set_file_contents(other_config_file("wireguard-ips"), json.dumps(wg_ips, indent=2))



"""Paracrine is a system deployment tool. It's based around [Mitogen](https://mitogen.networkgenomics.com/) and standard Python packages for
features (although, currently there's only the one core package). Current status is essentially,
"works for me, probably won't eat your computer", but thoughts and patches are welcomed.

It's designed towards rapid idempotent deploys i.e. a zero-changes deploy should take in the order of a small number of seconds ideally, and so can be used as your application deployment option as well as for system changes, which has the design benefit of you could use it for GitOps-style deploys. Obviously, if you do more changes, it'll take longer, but that's generally less of an issue, but it should still be faster than most other tooling options. It assumes it's got a host system to run from, which _probably_ can be anything that runs Python, but the testing for that so far has been on a Debian laptop.

It's named after [Paracrine signaling](https://en.wikipedia.org/wiki/Paracrine_signaling) "a type of cellular communication in which a cell produces a signal to induce changes in nearby cells" which feels pretty accurate for a deployment tool.

Usage
-----

1. Setup Python. Tested against 3.8+
2. `pip install paracrine`
3. Write a main file describing what you want to setup. [integration_test/main.py](https://github.com/palfrey/paracrine/blob/main/integration_test/main.py) is a reasonable example. It must call the `run` function, which takes arguments for the inventory file, and list of modules to run.
4. Write an inventory file for the machines this is managing. Current setup assumes they're all the same. [integration_test/docker/inventory.yaml](https://github.com/palfrey/paracrine/blob/main/integration_test/docker/inventory.yaml) is a reasonable example file, but I suggest generating it from whatever you're using to create the servers (e.g. Terraform).
5. Write a <code>config.yaml</code>. This has a main top-level key of `environments` with keys below that for each inventory file you've got ([integration_test/config.yaml](https://github.com/palfrey/paracrine/blob/main/integration_test/config.yaml) just has one, but in most scenarios you'll have at least a dev and prod setup). What you do below that is up to you, but typically it'll be environment variables and secrets to feed into the main file.
6. Run `python -m paracrine.commands.setup <inventory file>` - this will install the minimum python bits so that everything else works.
7. Run the main file (e.g. `python main.py -i ./docker/inventory.yaml`)

Utilities
---
All the utilities are callable as `python -m paracrine.commands.<utility> <inventory file>` and sometimes some extra args

* `paracrine.commands.login` - login to a specific server in the inventory. Index of which one is an additional arg
* `paracrine.commands.setup` - this will install the minimum python bits so that everything else works.

Built-in modules
---
* `paracrine.runners.aws` - Runs locally, gets the AWS access/secret keys of the current user during bootstrap for access later. Has a function `get_env_with_creds` for setting them in a run step using the `env` commmand.
* `paracrine.runners.certs` - Creates SSL certificates via LetsEncrypt. Takes two args "hostname" and "email".
* `paracrine.services.cockroachdb` - Sets up a [CockroachDB server](https://www.cockroachlabs.com/)
* `paracrine.services.pleroma` - Sets up a [Pleroma server](https://pleroma.social/). Uses the PostgreSQL and certs modules.
* `paracrine.services.postgresql` - Sets up a [PostgresSQL server](https://www.postgresql.org/)
* `paracrine.services.wireguard` - Sets up [Wireguard](https://www.wireguard.com/). Assumes your inventory template has a unique `wireguard_ip` line per server.

Writing a module file
---

"modules" are just Python files (or packages containing them). They should have one or more specially named functions - all of these are optional, but a module with none of these won't do anything.

* `dependencies` - Returns a list of modules this module requires (e.g. a database setup). Return type is `paracrine.deps.Modules` i.e. a list of either `ModuleType` or a Tuple of `ModuleType` and a value to be inserted as the `options` dictionary into the module.
* `local` - Function to be run locally i.e. on the running host before connecting to the destination machine
* `run` - Main function to run on the destination machine
* `parse_return` - Function to run locally with an argument of the result of running `run`

You can just use plain Python code, but `paracrine.helpers.fs/config/debian/network/python/systemd/users` have lots of useful functions you should preferably use instead.
"""

import os
import pathlib

DRY_RUN_ENV = "PARACRINE_DRY_RUN"


def is_dry_run() -> bool:
    return os.environ[DRY_RUN_ENV].lower() == "true"


def dry_run_safe_read(filepath: pathlib.Path, fakecontent: str) -> str:
    try:
        return filepath.open().read()
    except FileNotFoundError:
        if is_dry_run():
            return fakecontent
        raise



from typing import Dict

from .config import ServerDict, host, network_config, other_config, servers


def networks_by_interface(h):
    return dict([(intf["ifname"], intf) for intf in network_config(h["name"])])


def get_ipv4(network):
    addrs = network["addr_info"]
    for addr in addrs:
        if addr["family"] == "inet" and addr["local"] != "":
            return addr["local"]
    return None


def external_ip(h: ServerDict) -> str:
    return other_config(h["name"])["external_ip"]


def external_ips() -> Dict[str, object]:
    return dict([(h["name"], external_ip(h)) for h in servers()])


def wireguard_ips() -> Dict[str, str]:
    return dict(
        [(h["name"], h["wireguard_ip"]) for h in servers() if "wireguard_ip" in h]
    )


def wireguard_ip():
    return host()["wireguard_ip"]


def ips_for_network(net):
    return [
        x for x in [get_ipv4(intf) for (name, intf) in net.items()] if x is not None
    ]


def ips(h):
    network = networks_by_interface(h)
    return ips_for_network(network)



import importlib
from types import ModuleType
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

from mergedeep import merge

from .helpers.config import clear_return_data, get_return_data, set_data

Module = Union[ModuleType, Tuple[ModuleType, Dict[object, object]]]
Modules = Sequence[Module]
"""Type of modules handed to `paracrine.runner.run`"""

TransmitModule = Union[str, Tuple[str, Dict[object, object]]]
TransmitModules = Sequence[TransmitModule]


def runfunc(
    modules: Modules,
    name: str,
    arguments: Dict[str, Any] = {},
    data: Dict[str, object] = {},
) -> Dict[str, Any]:
    ret = {}

    def run(module: ModuleType, options: Dict[object, object]):
        func: Optional[Callable] = getattr(module, name, None)
        if func is not None:
            setattr(module, "options", options)
            clear_return_data()
            if data != {}:
                set_data(data)
            try:
                if module.__name__ not in ret:
                    ret[module.__name__] = []
                if module.__name__ in arguments:
                    info = func(arguments[module.__name__])
                else:
                    info = func()
                if isinstance(info, Dict):
                    info = merge({}, info, get_return_data())
                elif info is None:
                    info = get_return_data()
                ret[module.__name__].append(info)
            except Exception:
                print(f"Error while running {name} for {module.__name__}")
                raise

    for module in modules:
        if isinstance(module, ModuleType):
            run(module, {})
        else:
            run(module[0], module[1])
    return ret


def maketransmit_single(module: Module) -> TransmitModule:
    if isinstance(module, ModuleType):
        return module.__name__
    else:
        return (module[0].__name__, module[1])


def maketransmit(modules: Modules) -> TransmitModules:
    return [maketransmit_single(module) for module in modules]


def makereal(modules: TransmitModules) -> Modules:
    ret = []
    for module in modules:
        if isinstance(module, str):
            ret.append(importlib.import_module(module))
        else:
            ret.append((importlib.import_module(module[0]), module[1]))

    return ret



from ...helpers.config import in_docker
from ...helpers.fs import run_with_marker
from ...runners.core import use_this_host
from . import certs, node
from .common import CERTS_DIR, HOME_DIR, cockroach_binary, local_node_ip

options = {}


def dependencies():
    return [(node, options), certs]


def run():
    # FIXME: Can't make node start work in docker
    if in_docker() or not use_this_host("cockroach-init"):
        return
    COCKROACH_PORT = options.get("COCKROACH_PORT", 26257)
    run_with_marker(
        HOME_DIR.joinpath("init_done"),
        f"{cockroach_binary} init --certs-dir={CERTS_DIR} --host={local_node_ip()}:{COCKROACH_PORT}",
        run_if_command_changed=False,
    )
